@@ -5,12 +5,15 @@ import os
 import signal
 import unittest
 from itertools import product
+from textwrap import dedent
+from configparser import ConfigParser
 from unittest import mock
 from urllib.parse import urlparse
 
 import pytest
 
-from buku import DELIM, FIELD_FILTER, ALL_FIELDS, FetchResult, is_int, prep_tag_search, print_rec_with_filter, parse_range
+from buku import DELIM, FIELD_FILTER, ALL_FIELDS, SortKey, FetchResult, is_int, prep_tag_search, \
+                 print_rec_with_filter, get_netloc, extract_auth, parse_range, split_by_marker
 
 
 def check_import_html_results_contains(result, expected_result):
@@ -23,20 +26,58 @@ def check_import_html_results_contains(result, expected_result):
     return count == n * (n + 1) / 2
 
 
-@pytest.mark.parametrize(
-    "url, exp_res",
-    [
-        ["http://example.com", False],
-        ["ftp://ftp.somedomain.org", False],
-        ["http://examplecom.", True],
-        ["http://.example.com", True],
-        ["http://example.com.", True],
-        ["about:newtab", True],
-        ["chrome://version/", True],
-    ],
-)
+@pytest.mark.parametrize('url, result', [
+    ('http://user:password@hostname:1234/path?query#hash', 'user:password'),
+    ('http://:password@hostname:1234/path?query#hash', ':password'),
+    ('http://user:@hostname:1234/path?query#hash', 'user:'),
+    ('http://user@hostname:1234/path?query#hash', 'user'),
+    ('http://@hostname:1234/path?query#hash', ''),
+    ('http://hostname:1234/path?query#hash', None),
+    ('//[', ValueError('Invalid IPv6 URL')),
+    ('//⁈', ValueError("netloc '⁈' contains invalid characters under NFKC normalization")),
+])
+def test_extract_auth(url, result):
+    if not isinstance(result, Exception):
+        assert extract_auth(url) == (result, 'http://hostname:1234/path?query#hash')
+    else:
+        try:
+            extract_auth(url)
+        except Exception as e:
+            assert repr(e) == repr(result)
+        else:
+            assert False, f'expected {repr(result)} to be raised'
+
+
+@pytest.mark.parametrize('url, netloc', [
+    ['http://example.com', 'example.com'],
+    ['example.com/#foo/bar', 'example.com'],
+    ['ftp://ftp.somedomain.org', 'ftp.somedomain.org'],
+    ['about:newtab', None],
+    ['chrome://version/', 'version'],
+    ['javascript:void(0.0)', None],
+    ['data:,text.with.dots', None],
+    ['http://[', None],  # parsing error
+    ['http://⁈', None],  # parsing error
+])
+def test_get_netloc(url, netloc):
+    assert get_netloc(url) == netloc
+
+
+@pytest.mark.parametrize('url, exp_res', [
+    ['http://example.com', False],
+    ['example.com/#foo/bar', False],
+    ['ftp://ftp.somedomain.org', False],
+    ['http://examplecom.', True],   # ends with a '.'
+    ['http://.example.com', True],  # starts with a '.'
+    ['http://example.com.', True],  # ends with a '.'
+    ['about:newtab', True],
+    ['chrome://version/', True],    # contains no '.'
+    ['javascript:void(0.0)', True],
+    ['data:,text.with.dots', True],
+    ['http://[', True],             # parsing error
+    ['http://⁈', True],             # parsing error
+])
 def test_is_bad_url(url, exp_res):
-    """test func."""
     import buku
 
     res = buku.is_bad_url(url)
@@ -87,9 +128,9 @@ def test_get_PoolManager(m_myproxy):
 @pytest.mark.parametrize(
     "keywords, exp_res",
     [
-        ("", DELIM),
-        (",", DELIM),
-        ("tag1, tag2", ",t a g 1,t a g 2,"),
+        ([""], DELIM),
+        ([","], DELIM),
+        (["tag1, tag2"], ",tag1,tag2,"),
         ([" a tag , ,   ,  ,\t,\n,\r,\x0b,\x0c"], ",a tag,"),  # whitespaces
         ([",,,,,"], ","),  # empty tags
         (["\"tag\",'tag',tag"], ",\"tag\",'tag',tag,"),  # escaping quotes
@@ -114,14 +155,17 @@ def test_get_PoolManager(m_myproxy):
         ),
     ],
 )
-def test_parse_tags(keywords, exp_res):
+@pytest.mark.parametrize('prefix', [None, '', '+', '-'])
+def test_parse_tags(prefix, keywords, exp_res):
     """test func."""
     import buku
 
+    edit_input = prefix is not None
     if keywords is None:
-        assert buku.parse_tags(keywords) is None
+        assert buku.parse_tags(keywords, edit_input=edit_input) is None
     else:
-        assert buku.parse_tags(keywords) == exp_res
+        _keywords = ([] if not prefix else [prefix]) + keywords
+        assert buku.parse_tags(_keywords, edit_input=edit_input) == (prefix or '') + exp_res
 
 
 def test_parse_tags_no_args():
@@ -547,58 +591,62 @@ def test_is_nongeneric_url(url, exp_res):
     assert res == exp_res
 
 
-@pytest.mark.parametrize(
-    "newtag, exp_res",
-    [
-        (None, ("http://example.com", "text1", ",", None, 0, True, False)),
-        ("tag1", ("http://example.com", "text1", ",tag1,", None, 0, True, False)),
-    ],
-)
-def test_import_md(tmpdir, newtag, exp_res):
+@pytest.mark.parametrize('tags, newtag, exp_tags', [
+    ('', None, ','), ('', 'new tag', ',new tag,'),
+    ('foo, bar, baz', None, ',bar,baz,foo,'),
+    ('foo, bar, baz', 'new tag', ',bar,baz,foo,new tag,'),
+])
+@pytest.mark.parametrize('title', ['Bookmark title', '', None])
+@pytest.mark.parametrize('url', ['http://example.com', 'javascript:void(0)', ''])
+def test_import_md(tmpdir, url, title, tags, newtag, exp_tags):
     from buku import import_md
 
     p = tmpdir.mkdir("importmd").join("test.md")
-    p.write("[text1](http://example.com)")
+    print(line := (f'<{url}>' if title is None else f'[{title}]({url})') +
+                  ('' if not tags else f' <!-- TAGS: {tags} -->'))
+    p.write(line)
     res = list(import_md(p.strpath, newtag))
+    assert res == ([] if not url else  # `<>` and `[title]()` are not valid
+                   [(url, title or '', exp_tags, None, 0, True, False)])
+
+@pytest.mark.parametrize('newtag, exp_res', [
+    (None, ('http://example.com', 'text1', ',', None, 0, True, False)),
+    ('tag1', ('http://example.com', 'text1', ',tag1,', None, 0, True, False)),
+])
+@pytest.mark.parametrize('extension', ['.rss', '.atom'])
+def test_import_rss(tmpdir, extension, newtag, exp_res):
+    from buku import import_rss
+
+    p = tmpdir.mkdir('importrss').join('test' + extension)
+    p.write(
+        '<feed xmlns="http://www.w3.org/2005/Atom">\n'
+        '    <title>Bookmarks</title>\n'
+        '    <generator uri="https://github.com/jarun/buku">buku</generator>\n'
+        '    <entry>\n'
+        '        <title>text1</title>\n'
+        '        <link href="http://example.com"/>\n'
+        '    </entry>\n'
+        '</feed>\n')
+    res = list(import_rss(p.strpath, newtag))
     assert res[0] == exp_res
 
-
-@pytest.mark.parametrize(
-    "newtag, exp_res",
-    [
-        (
-            None,
-            (
-                "http://example.com",
-                "text1",
-                ",tag1,:tag2,tag:3,tag4:,tag::5,tag:6:,",
-                None,
-                0,
-                True,
-                False,
-            ),
-        ),
-        (
-            "tag0",
-            (
-                "http://example.com",
-                "text1",
-                ",tag0,tag1,:tag2,tag:3,tag4:,tag::5,tag:6:,",
-                None,
-                0,
-                True,
-                False,
-            ),
-        ),
-    ],
-)
-def test_import_org(tmpdir, newtag, exp_res):
+@pytest.mark.parametrize('tags, newtag, exp_tags', [
+    ('', None, ','), ('', 'new tag', ',new tag,'),
+    ('tag1: ::tag2:tag::3:tag4:: :tag:::5: ta g::6:: ', None, ',tag1,:tag2,tag:3,tag4:,tag::5,ta g:6:,'),
+    ('tag1: ::tag2:tag::3:tag4:: :tag:::5: ta g::6:: ', 'new tag', ',new tag,tag1,:tag2,tag:3,tag4:,tag::5,ta g:6:,'),
+])
+@pytest.mark.parametrize('title', ['Bookmark title', '', None])
+@pytest.mark.parametrize('url', ['http://example.com', 'javascript:void(0)', ''])
+def test_import_org(tmpdir, url, title, tags, newtag, exp_tags):
     from buku import import_org
 
     p = tmpdir.mkdir("importorg").join("test.org")
-    p.write("[[http://example.com][text1]] :tag1: ::tag2:tag::3:tag4:: :tag:::5:tag::6:: :")
+    print(line := (f'[[{url}]]' if title is None else f'[[{url}][{title}]]') +
+                  ('' if not tags else f' :{tags}:'))
+    p.write(line)
     res = list(import_org(p.strpath, newtag))
-    assert res[0] == exp_res
+    assert res == ([] if not url or title == '' else  # `[[]]`, `[[][title]]` and `[[url][]]` are not valid
+                   [(url, title or '', exp_tags, None, 0, True, False)])
 
 
 @pytest.mark.parametrize(
@@ -741,7 +789,7 @@ def test_import_html_and_add_parent():
             [
                 ("http://example11.com", None, ",folder11,", None, 0, True, False),
                 ("http://example121.com", None, ",folder121,", None, 0, True, False),
-                ("http://example12.com", None, None, None, 0, True, False),
+                ("http://example12.com", None, ",folder12,", None, 0, True, False),
                 ("http://example13.com", None, ",folder13 (blah blah),tag3,tag4,", None, 0, True, False),
             ],
         ),
@@ -802,6 +850,80 @@ def test_import_html_and_new_tag():
     assert res[0] == exp_res
 
 
+@pytest.mark.parametrize('profiles, expected', [
+    (dedent('''
+    [Profile3]
+    Name=ABCD
+    IsRelative=0
+    Path=/path/to/removable/drive/ABCD
+    Default=1
+
+    [Install4F96D1932A9F858E]
+    Default=/path/to/custom/path/my-main-profile
+    Locked=1
+
+    [Profile1]
+    Name=Main
+    IsRelative=0
+    Path=/path/to/custom/path/main-profile
+
+    [Profile0]
+    Name=default
+    IsRelative=1
+    Path=zsq8tck1.default-release
+
+    [InstallD087BC9767A4CB84]
+    Default=1koqf71l.default-nightly
+    Locked=1
+
+    [General]
+    StartWithLastProfile=1
+    Version=2
+    '''), ['/path/to/custom/path/my-main-profile', '1koqf71l.default-nightly']),
+    (dedent('''
+    [Profile3]
+    Name=ABCD
+    IsRelative=0
+    Path=/path/to/removable/drive/ABCD
+    Default=1
+
+    [Profile1]
+    Name=Main
+    IsRelative=0
+    Path=/path/to/custom/path/my-main-profile
+
+    [Profile0]
+    Name=default
+    IsRelative=1
+    Path=zsq8tck1.default-release
+
+    [General]
+    StartWithLastProfile=1
+    Version=2
+    '''), ['/path/to/removable/drive/ABCD', 'zsq8tck1.default-release']),
+    ('', []), (None, []),
+])
+@mock.patch('os.path.exists')
+def test_get_firefox_profile_names(_os_path_exists, profiles, expected):
+    _os_path_exists.return_value = profiles is not None
+    with mock.patch.object(ConfigParser, 'read', lambda self, _: self.read_string(profiles)):
+        import buku
+        assert buku.get_firefox_profile_names('') == expected
+
+@pytest.mark.parametrize('profiles, specified, expected', [
+    (['foo', '/bar/baz'], None, {
+        'foo': os.path.join('~/profiles', 'foo', 'places.sqlite'),
+        '/bar/baz': os.path.join('/bar/baz', 'places.sqlite'),
+    }),
+    (['foo', '/bar/baz'], 'qux', {'qux': os.path.join('~/profiles', 'qux', 'places.sqlite')}),
+    ([], '/grue/xyzzy', {'/grue/xyzzy': os.path.join('/grue/xyzzy', 'places.sqlite')}),
+])
+def test_get_firefox_db_paths(profiles, specified, expected):
+    with mock.patch('buku.get_firefox_profile_names', return_value=profiles):
+        import buku
+        assert buku.get_firefox_db_paths('~/profiles', specified) == expected
+
+
 @pytest.mark.parametrize(
     "platform, params",
     [
@@ -842,48 +964,75 @@ def test_copy_to_clipboard(platform, params):
 @pytest.mark.parametrize(
     "export_type, exp_res",
     [
-        [
-            "html",
-            "<!DOCTYPE NETSCAPE-Bookmark-file-1>\n\n"
-            '<META HTTP-EQUIV="Content-Type" CONTENT="text/html; charset=UTF-8">\n'
-            "<TITLE>Bookmarks</TITLE>\n<H1>Bookmarks</H1>\n\n<DL><p>\n"
-            '    <DT><H3 ADD_DATE="1556430615" LAST_MODIFIED="1556430615" PERSONAL_TOOLBAR_FOLDER="true">buku bookmarks</H3>\n'
-            "    <DL><p>\n"
-            '        <DT><A HREF="htttp://example.com" ADD_DATE="1556430615" LAST_MODIFIED="1556430615"></A>\n'
-            '        <DT><A HREF="htttp://example.org" ADD_DATE="1556430615" LAST_MODIFIED="1556430615"></A>\n'
-            '        <DT><A HREF="http://google.com" ADD_DATE="1556430615" LAST_MODIFIED="1556430615">Google</A>\n'
-            "    </DL><p>\n</DL><p>",
-        ],
-        [
-            "org",
-            "* [[htttp://example.com][Untitled]]\n* [[htttp://example.org][Untitled]]\n* [[http://google.com][Google]]\n",
-        ],
-        [
-            "markdown",
-            "- [Untitled](htttp://example.com)\n- [Untitled](htttp://example.org)\n- [Google](http://google.com)\n",
-        ],
         ["random", None],
         [
-            "xbel",
-            "\n".join(
-                [
-                    '<?xml version="1.0" encoding="UTF-8"?>',
-                    '<!DOCTYPE xbel PUBLIC "+//IDN python.org//'
-                    'DTD XML Bookmark Exchange Language 1.0//EN//XML" '
-                    '"http://pyxml.sourceforge.net/topics/dtds/xbel.dtd">',
-                    '<xbel version="1.0">',
-                    '<bookmark href="htttp://example.com">',
-                    "<title></title>",
-                    "</bookmark>",
-                    '<bookmark href="htttp://example.org">',
-                    "<title></title>",
-                    "</bookmark>",
-                    '<bookmark href="http://google.com">',
-                    "<title>Google</title>",
-                    "</bookmark>",
-                    "</xbel>",
-                ]
-            ),
+            'html',
+            '<!DOCTYPE NETSCAPE-Bookmark-file-1>\n\n'
+            '<META HTTP-EQUIV="Content-Type" CONTENT="text/html; charset=UTF-8">\n'
+            '<TITLE>Bookmarks</TITLE>\n<H1>Bookmarks</H1>\n\n<DL><p>\n'
+            '    <DT><H3 ADD_DATE="1556430615" LAST_MODIFIED="1556430615" PERSONAL_TOOLBAR_FOLDER="true">buku bookmarks</H3>\n'
+            '    <DL><p>\n'
+            '        <DT><A HREF="http://example.com" ADD_DATE="1556430615" LAST_MODIFIED="1556430615"></A>\n'
+            '        <DT><A HREF="http://example.org" ADD_DATE="1556430615" LAST_MODIFIED="1556430615" TAGS="bar,baz,foo"></A>\n'
+            '        <DT><A HREF="http://google.com" ADD_DATE="1556430615" LAST_MODIFIED="1556430615" TAGS="bar,baz,foo">Google</A>\n'
+            '    </DL><p>\n</DL><p>',
+        ],
+        [
+            'org',
+            '* [[http://example.com]]\n'
+            '* [[http://example.org]] :bar:baz:foo:\n'
+            '* [[http://google.com][Google]] :bar:baz:foo:\n',
+        ],
+        [
+            'markdown',
+            '- <http://example.com>\n'
+            '- <http://example.org> <!-- TAGS: bar,baz,foo -->\n'
+            '- [Google](http://google.com) <!-- TAGS: bar,baz,foo -->\n',
+        ],
+        [
+            'rss',
+            '<feed xmlns="http://www.w3.org/2005/Atom">\n'
+            '    <title>Bookmarks</title>\n'
+            '    <generator uri="https://github.com/jarun/buku">buku</generator>\n'
+            '    <entry>\n'
+            '        <title></title>\n'
+            '        <link href="http://example.com" rel="alternate" type="text/html"/>\n'
+            '        <id>1</id>\n'
+            '    </entry>\n'
+            '    <entry>\n'
+            '        <title></title>\n'
+            '        <link href="http://example.org" rel="alternate" type="text/html"/>\n'
+            '        <id>2</id>\n'
+            '        <category term="bar"/>\n'
+            '        <category term="baz"/>\n'
+            '        <category term="foo"/>\n'
+            '    </entry>\n'
+            '    <entry>\n'
+            '        <title>Google</title>\n'
+            '        <link href="http://google.com" rel="alternate" type="text/html"/>\n'
+            '        <id>3</id>\n'
+            '        <category term="bar"/>\n'
+            '        <category term="baz"/>\n'
+            '        <category term="foo"/>\n'
+            '    </entry>\n'
+            '</feed>',
+        ],
+        [
+            'xbel',
+            '<?xml version="1.0" encoding="UTF-8"?>\n'
+            '<!DOCTYPE xbel PUBLIC "+//IDN python.org//DTD XML Bookmark Exchange Language 1.0//EN//XML"'
+            ' "http://pyxml.sourceforge.net/topics/dtds/xbel.dtd">\n\n'
+            '<xbel version="1.0">\n'
+            '    <bookmark href="http://example.com">\n'
+            '        <title></title>\n'
+            '    </bookmark>\n'
+            '    <bookmark href="http://example.org" TAGS="bar,baz,foo">\n'
+            '        <title></title>\n'
+            '    </bookmark>\n'
+            '    <bookmark href="http://google.com" TAGS="bar,baz,foo">\n'
+            '        <title>Google</title>\n'
+            '    </bookmark>\n'
+            '</xbel>'
         ],
     ],
 )
@@ -892,9 +1041,9 @@ def test_convert_bookmark_set(export_type, exp_res, monkeypatch):
     from buku import convert_bookmark_set
 
     bms = [
-        (1, "htttp://example.com", "", ",", "", 0),
-        (1, "htttp://example.org", None, ",", "", 0),
-        (2, "http://google.com", "Google", ",", "", 0),
+        (1, "http://example.com", "", ",", "", 0),
+        (2, "http://example.org", None, ",bar,baz,foo,", "", 0),
+        (3, "http://google.com", "Google", ",bar,baz,foo,", "", 0),
     ]
     if export_type == "random":
         with pytest.raises(AssertionError):
@@ -940,28 +1089,58 @@ def test_get_data_from_page(charset, mode):
         'charset': f'\n<meta charset="{charset}"/>',
         'content': f'\n<meta http-equiv="content-type" content="text/html; charset={charset}"/>',
     }.get(mode, '')
-    body = f'<html>\n\n<head>{meta}\n<title>{title}</title>\n</head>\n<body></body>\n\n</html>\n'
+    keywords = '<meta name="keywords" content="foo, bar   baz, quux"/>'
+    body = f'<html>\n\n<head>{meta}\n{keywords}\n<title>{title}</title>\n</head>\n<body></body>\n\n</html>\n'
     resp = HTTPResponse(body.encode(charset), headers)
-    parsed_title, desc, keywords = get_data_from_page(resp)
-    assert parsed_title == title
+    parsed_title, desc, tags = get_data_from_page(resp)
+    assert (parsed_title, tags) == (title, "foo,bar baz,quux")
 
 
-@pytest.mark.parametrize('tokens, valid, expected', [
-    (None, None, None),
-    ('404', None, {404}),
-    ('403,404', None, {403, 404}),
-    ({'400', '500'}, None, {400, 500}),
-    (('400-404', '500'), None, {400, 401, 402, 403, 404, 500}),
-    (['400-404', '500'], lambda x: x in range(400, 600), {400, 401, 402, 403, 404, 500}),
-    (['400-404', '300'], lambda x: x in range(400, 600), ValueError('Not a valid range')),
+@pytest.mark.parametrize('tokens, kwargs, expected', [
+    (None, {}, None),
+    ('404', {}, {404}),
+    ('403,404', {}, {403, 404}),
+    ({'400', '500'}, {}, {400, 500}),
+    (('400-404', '500'), {}, {400, 401, 402, 403, 404, 500}),
+    (['400-404', '500'], {'valid': lambda x: x in range(400, 600)}, {400, 401, 402, 403, 404, 500}),
+    (['400-404', '300'], {'valid': lambda x: x in range(400, 600)}, ValueError('Not a valid range')),
+    ('-3', {}, {-3}),
+    ('-3', {'maxidx': 10}, {8, 9, 10}),
+    ('-30', {'maxidx': 3}, {1, 2, 3}),
+    ('10-3', {'maxidx': 5}, {3, 4, 5}),
 ])
-def test_parse_range(tokens, valid, expected):
+def test_parse_range(tokens, kwargs, expected):
     if not isinstance(expected, Exception):
-        assert parse_range(tokens, valid) == expected
+        assert parse_range(tokens, **kwargs) == expected
     else:
         try:
-            parse_range(tokens, valid)
+            parse_range(tokens, **kwargs)
             assert False, 'error expected'
         except Exception as e:
             assert type(e) is type(expected)
             assert str(e) == str(expected)
+
+
+def test_split_by_marker():
+    search_string = (' global substring  .title substring :url substring :https '
+                     '> description substring #partial,tags: #,exact,tags, *another global substring ')
+    assert split_by_marker(search_string) == [
+        ' global substring', '.title substring', ':url substring', ':https',
+        '> description substring', '#partial,tags:', '#,exact,tags,', '*another global substring ',
+    ]
+
+
+def test_SortKey():
+    assert repr(SortKey('foo', ascending=True)) == "+'foo'"
+    assert repr(SortKey('bar', ascending=False)) == "-'bar'"
+    assert SortKey('foo', ascending=True) > SortKey('bar', ascending=True)
+    assert not SortKey('foo', ascending=True) > SortKey('foo', ascending=True)  # pylint: disable=unnecessary-negation
+    assert not SortKey('foo', ascending=True) < SortKey('foo', ascending=True)  # pylint: disable=unnecessary-negation
+    assert not SortKey('foo', ascending=True) < SortKey('bar', ascending=True)  # pylint: disable=unnecessary-negation
+    assert SortKey('foo', ascending=False) < SortKey('bar', ascending=False)
+    assert not SortKey('foo', ascending=False) < SortKey('foo', ascending=False)  # pylint: disable=unnecessary-negation
+    assert not SortKey('foo', ascending=False) > SortKey('foo', ascending=False)  # pylint: disable=unnecessary-negation
+    assert not SortKey('foo', ascending=False) > SortKey('bar', ascending=False)  # pylint: disable=unnecessary-negation
+
+    custom_order = lambda s: (SortKey(len(s), ascending=False), SortKey(s, ascending=True))
+    assert sorted(['foo', 'bar', 'baz', 'quux'], key=custom_order) == ['quux', 'bar', 'baz', 'foo']
