@@ -3,14 +3,16 @@
 resources: https://flask.palletsprojects.com/en/2.2.x/testing/
 """
 from argparse import Namespace
+from unittest import mock
 
 import pytest
 from flask import request
 from lxml import etree
+from werkzeug.datastructures import MultiDict
 
 from buku import BukuDb
 from bukuserver import server
-from bukuserver.views import BookmarkModelView, TagModelView
+from bukuserver.views import BookmarkModelView, TagModelView, filter_key
 from tests.util import mock_fetch, _add_rec
 
 
@@ -26,7 +28,7 @@ def app(dbfile):
     yield app
     # clean up / reset resources here
 
-def env_fixture(name, **kwargs):
+def env_fixture(name, **kwargs):  # place this fixture BEFORE app or its dependencies
     """Produces a fixture that mocks a test parameter directly in an env var (before app init)"""
     def _env_fixture(dbfile, monkeypatch, request):
         if request.param is not None:  # default value placeholder
@@ -60,6 +62,12 @@ def bmv_instance(bukudb):
     return BookmarkModelView(bukudb)
 
 
+@pytest.mark.parametrize('idx, char', [('', ''), (0, '0'), (9, '9'), (10, 'A'), (35, 'Z'), (36, 'a'), (61, 'z')])
+def test_filter_key(idx, char):
+    with mock.patch('bukuserver.views.BookmarkModelView._filter_arg', return_value='filter_name'):
+        assert filter_key(None, idx) == f'flt{char}_filter_name'
+
+
 @pytest.mark.parametrize('disable_favicon', [False, True])
 def test_bookmark_model_view(bukudb, disable_favicon, app):
     inst = BookmarkModelView(bukudb)
@@ -98,7 +106,7 @@ def test_tag_model_view_get_list(tmv_instance, sort_field, sort_desc, filters, e
 ])
 def test_bmv_create_form(bmv_instance, url, backlink, app):
     with app.test_request_context():
-        request.args = {'link': url, 'url': backlink} if backlink else {'link': url}
+        request.args = MultiDict({'link': url, 'url': backlink} if backlink else {'link': url})
         form = bmv_instance.create_form()
         assert form.url.data == url
 
@@ -108,6 +116,7 @@ def test_bmv_create_form(bmv_instance, url, backlink, app):
 #
 
 xpath_alert = lambda kind, message: f'//div[@class="alert alert-{kind} alert-dismissable"][contains(., "{message}")]'
+xpath_cls = lambda s: ''.join(f'[contains(concat(" ", @class, " "), " {s} ")]' for s in s.split(' ') if s)
 
 def assert_success_alert(dom, edit, id=1):
     message = f'Record was successfully {"saved" if edit else "created"}.'
@@ -137,8 +146,8 @@ def assert_bookmark(bookmark, query, tags=None):
 @pytest.mark.gui
 @pytest.mark.slow
 @pytest.mark.parametrize('exists, uri, tab, args', [
-    (False, '/bookmark/new/', 'Create', ['link', 'title', 'description']),
-    (True, '/bookmark/edit/', 'Edit', ['id']),
+    (False, '/bookmark/new/', 'Create', ['link', 'title', 'description', 'popup']),
+    (True, '/bookmark/edit/', 'Edit', ['id', 'popup']),
 ])
 def test_bookmarklet_view(bukudb, client, exists, uri, tab, args):
     query = {'url': 'http://example.com', 'title': 'Sample site', 'description': 'Foo bar baz'}
@@ -147,7 +156,7 @@ def test_bookmarklet_view(bukudb, client, exists, uri, tab, args):
 
     response = client.get('/bookmarklet', query_string=query, follow_redirects=True)
     dom = assert_response(response, uri, argnames=args)
-    assert dom.xpath('//ul[@class="nav nav-tabs"]/li[@class="active"]/a/text()') == [tab]
+    assert dom.xpath(f'//ul{xpath_cls("nav nav-tabs")}/li{xpath_cls("active")}/a/text()') == [tab]
     assert dom.xpath('//input[@name="link"]/@value') == [query['url']]
     assert bool(dom.xpath('//input[@name="id"]')) == exists
 
@@ -159,12 +168,14 @@ def test_bookmarklet_view(bukudb, client, exists, uri, tab, args):
     (True, '', 'Some description'),
     (False, 'Some title', ''),
     (False, '', 'Some description'),
+    (None, 'Some title', ''),
+    (None, '', 'Some description'),
 ])
 def test_create_and_fetch(bukudb, monkeypatch, client, fetch, title, desc):
     query = {'link': 'http://example.com', 'title': title, 'description': desc, 'tags': 'foo, bar, baz'}
-    _title, _desc = 'Sample site', 'Foo bar baz'
-    if fetch:
-        query['fetch'] = 'on'
+    _title, _desc = 'Fetched title', 'Fetched description'
+    if fetch is not None:
+        query['fetch'] = 'on' if fetch else ''
 
     with mock_fetch(title=_title, desc=_desc):
         response = client.post('/bookmark/new/', data=query, follow_redirects=True)
@@ -173,8 +184,8 @@ def test_create_and_fetch(bukudb, monkeypatch, client, fetch, title, desc):
     [bookmark] = bukudb.get_rec_all()
     assert_bookmark(bookmark, {
         'link': query['link'], 'tags': ',bar,baz,foo,',
-        'title': (title or _title) if fetch else title,
-        'description': (desc or _desc) if fetch else desc,
+        'title': (title or _title) if fetch or fetch is None else title,  # defaults to True
+        'description': (desc or _desc) if fetch or fetch is None else desc,
     })
 
 
@@ -185,7 +196,7 @@ def test_create_and_fetch(bukudb, monkeypatch, client, fetch, title, desc):
     ('_continue_editing', '/bookmark/edit/', {'id': '1', 'url': '/bookmark/'}),
 ])
 def test_create_redirect(client, redirect, uri, args):
-    query = {'link': 'http://example.com', 'title': '', 'description': '', 'tags': '', redirect: 'on'}
+    query = {'link': 'http://example.com', 'title': '', 'description': '', 'tags': '', 'fetch': '', redirect: 'on'}
 
     response = client.post('/bookmark/new/', data=query, follow_redirects=True)
     dom = assert_response(response, uri, args=args)
@@ -278,7 +289,7 @@ def test_env_per_page(bukudb, app, client, total, per_page, pages, last_page):
 
     response = client.get('/bookmark/last-page', follow_redirects=True)
     dom = assert_response(response, '/bookmark/', args={'page': str(pages - 1)})
-    cells = dom.xpath('//td[@class="col-Entry"]')
+    cells = dom.xpath(f'//td{xpath_cls("col-entry")}')
     assert len(cells) == last_page
     for i, cell in enumerate(cells, start=1):
         url = f'http://example.com/{total - last_page + i}'
@@ -289,9 +300,21 @@ def test_env_per_page(bukudb, app, client, total, per_page, pages, last_page):
 @pytest.mark.slow
 @pytest.mark.parametrize('new_tab', [False, True, None])
 @pytest.mark.parametrize('favicons', [False, True, None])
-@pytest.mark.parametrize('mode', ['full', 'netloc', None])
+@pytest.mark.parametrize('mode', ['full', 'netloc', 'netloc-tag', None])
 def test_env_entry_render_params(bukudb, app, client, mode, favicons, new_tab):
-    url, netloc, title, desc, tags = 'http://example.com', 'example.com', 'Sample site', 'Foo bar baz', ',bar,baz,foo,'
+    _test_env_entry_render_params(bukudb, app, client, mode, favicons, new_tab, 'http://example.com', 'example.com', 'Sample site')
+
+@pytest.mark.parametrize('url, netloc, title', [
+    ('http://example.com', 'example.com', ''),
+    ('javascript:void(0)', '', 'Sample site'),
+    ('javascript:void(0)', '', ''),
+])
+@pytest.mark.parametrize('mode', ['full', 'netloc', 'netloc-tag'])
+def test_env_entry_render_params_blanks(bukudb, app, client, mode, url, netloc, title):
+    _test_env_entry_render_params(bukudb, app, client, mode, True, True, url, netloc, title)
+
+def _test_env_entry_render_params(bukudb, app, client, mode, favicons, new_tab, url, netloc, title):
+    desc, tags = 'Foo bar baz', ',bar,baz,foo,'
     _add_rec(bukudb, url, title, tags, desc)
     _tags = tags.strip(',').split(',')
     if mode:
@@ -302,18 +325,23 @@ def test_env_entry_render_params(bukudb, app, client, mode, favicons, new_tab):
         app.config.update({'BUKUSERVER_OPEN_IN_NEW_TAB': new_tab})
 
     dom = assert_response(client.get('/bookmark/'), '/bookmark/')
-    cell = ' '.join(etree.tostring(dom.xpath('//td[@class="col-Entry"]')[0], encoding='unicode').strip().split())
+    cell = ' '.join(etree.tostring(dom.xpath(f'//td{xpath_cls("col-entry")}')[0], encoding='unicode').strip().split())
     target = '' if not new_tab else ' target="_blank"'
-    icon = '' if not favicons else f'<img src="http://www.google.com/s2/favicons?domain={netloc}"/> '
-    prefix = f'<td class="col-Entry"> {icon}<a href="{url}"{target}>{title}</a>'
-    suffix = (''.join(f'<a class="btn btn-default" href="/bookmark/?flt0_tags_contain={s}">{s}</a>' for s in _tags) +
-              f'<br/>{desc} </td>')
+    icon = '' if not favicons else (netloc and f'<img class="favicon" src="http://www.google.com/s2/favicons?domain={netloc}"/> ')
+    urltext = title or '&lt;EMPTY TITLE&gt;'
+    _title = (urltext if not netloc and mode in ('full', None) else f'<a href="{url}"{target}>{urltext}</a>')
+    prefix = f'<td class="col-entry"> {icon}<span class="title" title="{url}">{_title}</span>'
+    tags = [f'<a class="btn label label-default" href="/bookmark/?flt0_tags_contain={s}">{s}</a>' for s in _tags]
+    netloc_tag = ('' if mode == 'netloc' or not netloc else
+                  f'<a class="btn label label-success" href="/bookmark/?flt0_url_netloc_match={netloc}">netloc:{netloc}</a>')
+    suffix = f'<div class="tag-list">{netloc_tag}{"".join(tags)}</div><div class="description">{desc}</div> </td>'
     if mode == 'netloc':
-        assert cell == prefix + f' (<a href="/bookmark/?flt0_url_netloc_match={netloc}">{netloc}</a>)<br/>' + suffix
+        _netloc = netloc and f'<span class="netloc"> (<a href="/bookmark/?flt0_url_netloc_match={netloc}">{netloc}</a>)</span>'
+        assert cell == prefix + _netloc + suffix
+    elif mode == 'netloc-tag':
+        assert cell == prefix + suffix
     else:
-        assert cell == (prefix + f'<br/><a href="{url}"{target}>{url}</a><br/>' +
-                        f'<a class="btn btn-default" href="/bookmark/?flt0_url_netloc_match={netloc}">netloc:{netloc}</a>' +
-                        suffix)
+        assert cell == f'{prefix}<span class="link"><a href="{url}"{target}>{url}</a></span>{suffix}'
 
 
 readonly = env_fixture('BUKUSERVER_READONLY', params=[False, True, None])
@@ -326,12 +354,12 @@ def test_env_readonly(bukudb, readonly, client):
 
     response = client.get('/bookmark/')
     dom = assert_response(response, '/bookmark/')
-    assert bool(dom.xpath('//td[@class="list-buttons-column"]/a[@title="Edit Record"]')) == edit, 'edit icon'
-    assert bool(dom.xpath('//td[@class="list-buttons-column"]/form[@action="/bookmark/delete/"]')) == edit, 'delete icon'
+    assert bool(dom.xpath(f'//td{xpath_cls("list-buttons-column")}/a[@title="Edit Record"]')) == edit, 'edit icon'
+    assert bool(dom.xpath(f'//td{xpath_cls("list-buttons-column")}/form[@action="/bookmark/delete/"]')) == edit, 'delete icon'
 
     response = client.get('/bookmark/details/', query_string={'id': 1})
     dom = assert_response(response, '/bookmark/details/')
-    assert (dom.xpath('//ul[@class="nav nav-tabs"]/li/a/text()') ==
+    assert (dom.xpath(f'//ul{xpath_cls("nav nav-tabs")}/li/a/text()') ==
             (['List', 'Details'] if readonly else ['List', 'Create', 'Edit', 'Details']))
 
     response = client.get('/bookmark/new/', follow_redirects=True)
@@ -348,7 +376,7 @@ def test_env_reverse_proxy_path(proxy_path, client):
     links = [(proxy_path or '') + s for s in ['/', '/bookmark/', '/tag/', '/statistic/']]
 
     dom = assert_response(client.get(links[0]), links[0])
-    assert dom.xpath('//nav[@class="navbar navbar-default"]//a/@href') == ['/'] + links
+    assert dom.xpath(f'//nav{xpath_cls("navbar")}//a/@href ') == ['/'] + links
     body_links = dom.xpath('//main//a/@href')
     assert body_links[-1].startswith('javascript:')
     assert body_links[:-1] == links[1:]
@@ -364,37 +392,35 @@ theme = env_fixture('BUKUSERVER_THEME', params=['default', 'slate', None])
 @pytest.mark.slow
 def test_env_theme(theme, client):
     dom = assert_response(client.get('/'), '/')
-    assert dom.xpath('//head/link[@rel="stylesheet"][@href=$href]',
-                     href=f'/static/admin/bootstrap/bootstrap3/swatch/{theme or "default"}/bootstrap.min.css?v=3.3.5')
+    assert dom.xpath('//head/link[@rel="stylesheet"][starts-with(@href, $href)]',
+                     href=f'/static/admin/bootstrap/bootstrap3/swatch/{theme or "default"}/bootstrap.min.css?')
 
 
 _DICT = {
-    'en': {'//ul[@class="nav navbar-nav"]/li/a/text()': ['Home', 'Bookmarks', 'Tags', 'Statistic'],
-           '//ul[@class="nav nav-tabs actions-nav"]/li/a/text()': ['List (1)', 'Create', 'Add Filter', '10 items'],
-           '//td[@class="list-buttons-column"]/a/@title': ['View Record', 'Edit Record'],
-           '//td[@class="list-buttons-column"]/form/button/@title': ['Delete record']},
-    'de': {'//ul[@class="nav navbar-nav"]/li/a/text()': ['Start', 'Bookmarks', 'Tags', 'Statistic'],
-           '//ul[@class="nav nav-tabs actions-nav"]/li/a/text()': ['Liste (1)', 'Erstellen', 'Filter hinzufügen', '10 Elemente'],
-           '//td[@class="list-buttons-column"]/a/@title': ['Eintrag ansehen', 'Eintrag bearbeiten'],
-           '//td[@class="list-buttons-column"]/form/button/@title': ['Delete record']},
-    'fr': {'//ul[@class="nav navbar-nav"]/li/a/text()': ['Accueil', 'Bookmarks', 'Tags', 'Statistic'],
-           '//ul[@class="nav nav-tabs actions-nav"]/li/a/text()': ['Liste (1)', 'Créer', 'Ajouter un filtre', '10 items'],
-           '//td[@class="list-buttons-column"]/a/@title': ['Afficher L\'enregistrement', 'Modifier enregistrement'],
-           '//td[@class="list-buttons-column"]/form/button/@title': ['Delete record']},
-    'ru': {'//ul[@class="nav navbar-nav"]/li/a/text()': ['Главная', 'Bookmarks', 'Tags', 'Statistic'],
-           '//ul[@class="nav nav-tabs actions-nav"]/li/a/text()': ['Список (1)', 'Создать', 'Добавить Фильтр', '10 элементы'],
-           '//td[@class="list-buttons-column"]/a/@title': ['Просмотр записи', 'Редактировать запись'],
-           '//td[@class="list-buttons-column"]/form/button/@title': ['Delete record']},
+    'en': {f'//ul{xpath_cls("nav navbar-nav")}/li/a/text()': ['Home', 'Bookmarks', 'Tags', 'Statistic'],
+           f'//ul{xpath_cls("nav nav-tabs")}/li/a/text()': ['List (1)', 'Create', 'Random', 'Add Filter', '10 items'],
+           f'//td{xpath_cls("list-buttons-column")}/a/@title': ['View Record', 'Edit Record'],
+           f'//td{xpath_cls("list-buttons-column")}/form/button/@title': ['Delete record']},
+    'de': {f'//ul{xpath_cls("nav navbar-nav")}/li/a/text()': ['Start', 'Lesezeichen', 'Schilder', 'Statistik'],
+           f'//ul{xpath_cls("nav nav-tabs")}/li/a/text()': ['Liste (1)', 'Erstellen', 'Zufälliger', 'Filter hinzufügen', '10 Elemente'],
+           f'//td{xpath_cls("list-buttons-column")}/a/@title': ['Eintrag ansehen', 'Eintrag bearbeiten'],
+           f'//td{xpath_cls("list-buttons-column")}/form/button/@title': ['Delete record']},  # ['Datensatz löschen']},
+    'fr': {f'//ul{xpath_cls("nav navbar-nav")}/li/a/text()': ['Accueil', 'Signets', 'Étiquettes', 'Statistique'],
+           f'//ul{xpath_cls("nav nav-tabs")}/li/a/text()': ['Liste (1)', 'Créer', 'Aléatoire', 'Ajouter un filtre', '10 items'],
+           f'//td{xpath_cls("list-buttons-column")}/a/@title': ['Afficher L\'enregistrement', 'Modifier enregistrement'],
+           f'//td{xpath_cls("list-buttons-column")}/form/button/@title': ['Delete record']},  # ['Supprimer l\'enregistrement']},
+    'ru': {f'//ul{xpath_cls("nav navbar-nav")}/li/a/text()': ['Главная', 'Закладки', 'Теги', 'Статистика'],
+           f'//ul{xpath_cls("nav nav-tabs")}/li/a/text()': ['Список (1)', 'Создать', 'Случайная', 'Добавить Фильтр', '10 элементы'],
+           f'//td{xpath_cls("list-buttons-column")}/a/@title': ['Просмотр записи', 'Редактировать запись'],
+           f'//td{xpath_cls("list-buttons-column")}/form/button/@title': ['Delete record']},  # ['Удалить запись']},
 }
+locale = env_fixture('BUKUSERVER_LOCALE', params=['en', 'de', 'fr', 'ru', None])
 
 @pytest.mark.gui
 @pytest.mark.slow
-@pytest.mark.parametrize('locale', ['en', 'de', 'fr', 'ru', None])
-def test_env_locale(bukudb, app, client, locale):
+def test_env_locale(bukudb, locale, client):
     strings = _DICT[locale or 'en']
     _add_rec(bukudb, 'http://example.com')
-    if locale:
-        app.config.update({'BUKUSERVER_LOCALE': locale})
 
     dom = assert_response(client.get('/bookmark/'), '/bookmark/')
     for k, v in strings.items():
